<?php

namespace Drupal\media_twig_tools\Config;

use Drupal\Core\Config\Config;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Entity\EntityStorageInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\File\FileSystemInterface;

/**
 * Configuration for MediaTwigTools settings.
 *
 * @package Drupal\media_twig_tools\Config
 */
class MediaTwigToolsConfig implements MediaTwigToolsConfigInterface {

  /**
   * Serivce name.
   *
   * @const string
   */
  public const SERVICE_NAME = 'media_twig_tools.config';

  /**
   * Field dirname.
   *
   * @const string
   */
  public const FIELD_DIRNAME = 'dirname';

  /**
   * Field no style.
   *
   * @const string
   */
  public const FIELD_NO_STYLE = 'no_style';

  /**
   * Default no style identifier.
   *
   * @const string
   */
  public const DEFAULT_NO_STYLE_IDENTIFIER = 'unstyle';

  /**
   * Allow max size field.
   *
   * @const string
   */
  public const FIELD_ALLOW_MAX_SIZE = 'allow_max_size';

  /**
   * Enable format bounce.
   *
   * @const string
   */
  public const FIELD_ENABLE_FORMAT_BOUNCE = 'enable_format_bounce';

  /**
   * Enable Oversize check.
   *
   * @const string
   */
  public const FIELD_ENABLE_OVERSIZE_CHECK = 'enable_oversize_check';

  /**
   * Config.
   *
   * @var \Drupal\Core\Config\Config
   */
  protected Config $conf;

  /**
   * The file System.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * File storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $fileStorage;

  /**
   * Image style storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected EntityStorageInterface $imageStyleStorage;

  /**
   * Directories list.
   *
   * @var array
   */
  protected array $dirs = [];

  /**
   * MediaTwigToolsConfig constructor.
   */
  public function __construct(ConfigFactoryInterface $configFactory, FileSystemInterface $fileSystem, EntityTypeManagerInterface $entityTypeManager) {
    $this->conf = $configFactory->getEditable(static::SERVICE_NAME);
    $this->fileSystem = $fileSystem;
    $this->fileStorage = $entityTypeManager->getStorage('file');
    $this->imageStyleStorage = $entityTypeManager->getStorage('image_style');
  }

  /**
   * Singleton quick access.
   *
   * @return static
   *   Singleton.
   */
  public static function me() {
    return \Drupal::service(static::SERVICE_NAME);
  }

  /**
   * {@inheritdoc}
   */
  public function existsNoStyleIdentifier($value) {
    return array_key_exists($value, $this->getImageStyleStorage()->loadMultiple());
  }

  /**
   * {@inheritdoc}
   */
  public function getDirname() {
    return $this->conf->get(static::FIELD_DIRNAME) ?: 'builder';
  }

  /**
   * {@inheritdoc}
   */
  public function setDirname($dirname) {
    $this->conf
      ->set(static::FIELD_DIRNAME, $dirname)
      ->save();

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getNoStyle() {
    return $this->conf->get(static::FIELD_NO_STYLE) ?: static::DEFAULT_NO_STYLE_IDENTIFIER;
  }

  /**
   * {@inheritdoc}
   */
  public function setNoStyle($noStyle) {
    $this->conf
      ->set(static::FIELD_NO_STYLE, $noStyle)
      ->save();

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function getDirs($force = FALSE) {
    if ($force || empty($this->dirs)) {
      $this->dirs = ['public' => [], 'private' => []];
      foreach ($this->dirs as $key => &$dir) {
        $dir['path'] = $this->fileSystem->realpath($key . '://' . $this->getDirname());
        try {
          $dir['files'] = glob($dir['path'] . '/*');
        }
        catch (\Exception $e) {
          $dir['files'] = [];
        }
      }
    }

    return $this->dirs;
  }

  /**
   * {@inheritdoc}
   */
  public function clearDir($dirname) {
    $this->fileSystem->deleteRecursive($dirname);
  }

  /**
   * {@inheritdoc}
   */
  public function getFileStorage(): EntityStorageInterface {
    return $this->fileStorage;
  }

  /**
   * {@inheritdoc}
   */
  public function getImageStyleStorage(): EntityStorageInterface {
    return $this->imageStyleStorage;
  }

  /**
   * {@inheritdoc}
   */
  public function getAllowMaxSize() {
    return $this->conf->get(static::FIELD_ALLOW_MAX_SIZE);
  }

  /**
   * {@inheritdoc}
   */
  public function setAllowMaxSize($allowMaxSize) {
    $this->conf
      ->set(static::FIELD_ALLOW_MAX_SIZE, $allowMaxSize)
      ->save();

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function isFormatBounceEnable() {
    return $this->conf->get(static::FIELD_ENABLE_FORMAT_BOUNCE) ?? FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function enableFormatBounce($enable = TRUE) {
    $this->conf
      ->set(static::FIELD_ENABLE_FORMAT_BOUNCE, $enable)
      ->save();

    return $this;
  }

  /**
   * {@inheritdoc}
   */
  public function isOversizeCheckEnable() {
    return $this->conf->get(static::FIELD_ENABLE_OVERSIZE_CHECK) ?? FALSE;
  }

  /**
   * {@inheritdoc}
   */
  public function enableOversizeCheck($enable = TRUE) {
    $this->conf
      ->set(static::FIELD_ENABLE_OVERSIZE_CHECK, $enable)
      ->save();

    return $this;
  }

}
