<?php

namespace Drupal\media_twig_tools\Services;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Field\EntityReferenceFieldItemList;
use Drupal\Core\Field\Plugin\Field\FieldType\EntityReferenceItem;
use Drupal\Core\File\FileSystemInterface;
use Drupal\Core\File\FileUrlGeneratorInterface;
use Drupal\Core\Language\LanguageManagerInterface;
use Drupal\Core\Render\Element;
use Drupal\file\FileInterface;
use Drupal\file\Plugin\Field\FieldType\FileFieldItemList;
use Drupal\image\ImageStyleInterface;
use Drupal\media\Entity\Media;
use Drupal\media\MediaInterface;
use Drupal\media_twig_tools\Config\MediaTwigToolsConfig;
use Twig\Extension\AbstractExtension;
use Twig\TwigFunction;

/**
 * Provides twig tools for facilitate media integration.
 *
 * - Render the image tag directly :
 *    {{ imgFromMedia(content.field_media_1, {style:'my_style', class:'class1
 * class2', title:'My title', {attributes: {'data-attr-1':'1']}}) }}
 *
 * @package Drupal\media_twig_tools\Services
 */
class MediaTwigTools extends AbstractExtension {

  /**
   * OPTION style.
   *
   * @const string
   */
  const OPTION_STYLE = 'style';

  /**
   * OPTION nom du champ.
   *
   * @const string
   */
  const OPTION_FIELD_NAME = 'field_name';

  /**
   * OPTION class.
   *
   * @const string
   */
  const OPTION_CLASS = 'class';

  /**
   * OPtion title.
   *
   * @const string
   */
  const OPTION_TITLE = 'title';

  /**
   * OPTION dimension.
   *
   * @const string
   */
  const OPTION_DIMENSION = 'dimension';

  /**
   * OPTION Settings.
   *
   * @const string
   */
  const OPTION_SETTINGS = 'settings';

  /**
   * OPTION relative url.
   *
   * @const strings
   */
  const OPTION_RELATIVE = 'relative';

  /**
   * OPTION attribute.
   *
   * @const string
   */
  const OPTION_ATTRIBUTES = 'attributes';

  /**
   * OPTION rules.
   *
   * @const string
   */
  const OPTION_RULES = 'rules';

  /**
   * OPTION no-size.
   *
   * @const string
   */
  const OPTION_NO_SIZE = 'no_size';

  /**
   * OPTION Override src.
   *
   * @const string
   */
  public const OPTION_OVERRIDE_SRC = 'data-override-src';

  /**
   * Service name.
   *
   * @const string
   */
  const SERVICE_NAME = 'media_twig_tools.media_twig_tools';

  /**
   * Entity TYpe manager.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * LanguageManager.
   *
   * @var \Drupal\Core\Language\LanguageManagerInterface
   */
  protected LanguageManagerInterface $languageManager;

  /**
   * ImageBuilder.
   *
   * @var \Drupal\media_twig_tools\Services\ImageBuilder
   */
  protected ImageBuilder $imageBuilder;

  /**
   * The configuration.
   *
   * @var \Drupal\media_twig_tools\Config\MediaTwigToolsConfig
   */
  protected MediaTwigToolsConfig $conf;

  /**
   * The file system.
   *
   * @var \Drupal\Core\File\FileSystemInterface
   */
  protected FileSystemInterface $fileSystem;

  /**
   * Check oversize.
   *
   * @var \Drupal\media_twig_tools\Services\OversizeChecker
   */
  protected OversizeChecker $oversizeChecker;

  /**
   * File Url Generator.
   *
   * @var \Drupal\Core\File\FileUrlGeneratorInterface
   */
  protected FileUrlGeneratorInterface $fileUrlGenerator;

  /**
   * TwigUtils constructor.
   *
   * @param \Drupal\Core\Entity\EntityTypeManagerInterface $entity_type_manager
   *   The entity type manager.
   * @param \Drupal\Core\Language\LanguageManagerInterface $language_manager
   *   The language manager.
   * @param \Drupal\media_twig_tools\Services\ImageBuilder $image_builder
   *   The image builder.
   * @param \Drupal\media_twig_tools\Config\MediaTwigToolsConfig $conf
   *   The configuration.
   * @param \Drupal\Core\File\FileSystemInterface $file_system
   *   The file system.
   * @param \Drupal\media_twig_tools\Services\OversizeChecker $oversize_checker
   *   The oversize checker.
   * @param \Drupal\Core\File\FileUrlGeneratorInterface $file_url_generator
   *   The file url generator.
   */
  public function __construct(
    EntityTypeManagerInterface $entity_type_manager,
    LanguageManagerInterface $language_manager,
    ImageBuilder $image_builder,
    MediaTwigToolsConfig $conf,
    FileSystemInterface $file_system,
    OversizeChecker $oversize_checker,
    FileUrlGeneratorInterface $file_url_generator,
  ) {
    $this->entityTypeManager = $entity_type_manager;
    $this->languageManager = $language_manager;
    $this->imageBuilder = $image_builder;
    $this->conf = $conf;
    $this->fileSystem = $file_system;
    $this->oversizeChecker = $oversize_checker;
    $this->fileUrlGenerator = $file_url_generator;
  }

  /**
   * Retourne le singleton.
   *
   * @return static
   *   Le singleton.
   */
  public static function me() {
    return \Drupal::service(static::SERVICE_NAME);
  }

  /**
   * {@inheritdoc}
   */
  public function getFunctions() {
    return [
      new TwigFunction('imgFromMedia', [$this, 'imgFromMedia']),
      new TwigFunction('getMediaUrl', [$this, 'getMediaUrl']),
      new TwigFunction('pictureFromMedia', [$this, 'pictureFromMedia']),
    ];
  }

  /**
   * Return the default options.
   *
   * @param array $options
   *   The user options.
   *
   * @return array
   *   THe merged options
   */
  public function getDefaultOptions(array $options) {
    $style = $options[static::OPTION_STYLE] ?? NULL;
    if ($style) {
      $style = $style instanceof ImageStyleInterface ? $style : $this->conf->getImageStyleStorage()
        ->load($style);
    }

    return [
      static::OPTION_STYLE => $style,
      static::OPTION_CLASS => $options[static::OPTION_CLASS] ?? NULL,
      static::OPTION_FIELD_NAME => $options[static::OPTION_FIELD_NAME] ?? 'field_media_image',
      static::OPTION_TITLE => $options[static::OPTION_TITLE] ?? NULL,
      static::OPTION_DIMENSION => isset($options[static::OPTION_DIMENSION]) ? TRUE : FALSE,
      static::OPTION_SETTINGS => $options[static::OPTION_SETTINGS] ?? [],
      static::OPTION_RELATIVE => $options[static::OPTION_RELATIVE] ?? TRUE,
      static::OPTION_ATTRIBUTES => $options[static::OPTION_ATTRIBUTES] ?? [],
      static::OPTION_NO_SIZE => $options[static::OPTION_NO_SIZE] ?? TRUE,
    ] + $options;
  }

  /**
   * Return an image build array from a media.
   *
   * @param mixed $media_params
   *   The media or the media build array.
   * @param array $options
   *   The options.
   *
   * @return array
   *   The image build array.
   */
  public function imgFromMedia($media_params, array $options = []): array {
    // Définitions des options par défaut.
    $options = $this->getDefaultOptions($options);

    // Pour chaque média, on retourne le build array.
    $build_array = [];
    foreach ($this->getMediaEntityFromTwigParam($media_params) as $media) {
      if ($build = $this->getImageBuildArrayFromMedia($media, $options)) {
        $build_array[] = $build;
      }
    }

    return $build_array;
  }

  /**
   * Return the list of media entities from a twig parameter.
   *
   * @param mixed $media_params
   *   The media params.
   *
   * @return array
   *   The media entity list.
   */
  public function getMediaEntityFromTwigParam($media_params): array {
    $media_list = [];
    if (is_array($media_params)) {
      // Parcours du tableau de média.
      foreach (Element::children($media_params) as $key) {
        if ($media_params[$key] instanceof Media) {
          // Liste de média brut.
          $media_list[] = $media_params[$key];
        }
        elseif ($media_params[$key]['#media'] instanceof Media) {
          // List de build array.
          $media_list[] = $media_params[$key]['#media'];
        }
      }
    }
    elseif ($media_params instanceof Media) {
      $media_list = [$media_params];
    }
    elseif ($media_params instanceof EntityReferenceFieldItemList) {
      $media_list = $media_params->referencedEntities();
    }
    elseif ($media_params instanceof EntityReferenceItem) {
      $media_list = [
        $this->entityTypeManager->getStorage('media')
          ->load($media_params->target_id),
      ];
    }

    // Manage translations.
    $langcode = $this->languageManager->getCurrentLanguage()->getId();

    return array_map(
      function (MediaInterface $media) use ($langcode) {
        return $media->hasTranslation($langcode) ? $media->getTranslation($langcode) : $media;
      },
      $media_list
    );
  }

  /**
   * Return the build array of the image formatter according to options.
   *
   * @param \Drupal\media\MediaInterface $media
   *   The media.
   * @param array $options
   *   The options.
   *
   * @return array|false
   *   The build array.
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  protected function getImageBuildArrayFromMedia(MediaInterface $media, array $options) {
    // Get the file.
    /** @var \Drupal\file\FileInterface $file */
    $file = $media->get($options[static::OPTION_FIELD_NAME])
      ->referencedEntities()[0];
    if (is_null($file)) {
      return FALSE;
    }

    // Get the render of the main item of the media.
    $formatter_build = $media->get($options[static::OPTION_FIELD_NAME])
      ->first()
      ->view();

    // Redefine meta data.
    $this->initImageMeta($formatter_build, $media, $options);

    // Manage image style.
    $this->initImageFormatterStyleData($formatter_build, $file, $options);

    // Manage classes.
    $this->initImagesClasses($formatter_build, $options);

    // Manage image tag.
    $this->initImageTag($formatter_build);

    // Init attributes.
    $this->initAttributes($formatter_build, $options);

    // Init srcset.
    $this->initSrcSet($formatter_build, $media, $options);

    // Init rules.
    $this->initRules($formatter_build, $media, $options);

    // No size.
    if ($options[static::OPTION_NO_SIZE]) {
      $formatter_build['#item_attributes'][static::OPTION_NO_SIZE] = 1;
    }

    return $formatter_build;
  }

  /**
   * Return the image style data.
   *
   * @param array $formatter_build
   *   The image formatter build array.
   * @param \Drupal\file\FileInterface $file
   *   The file.
   * @param array $options
   *   The options.
   */
  protected function initImageFormatterStyleData(array &$formatter_build, FileInterface $file, array $options) {
    // Récupération de l'url stylée.
    /** @var \Drupal\image\ImageStyleInterface $style */
    $style = $options[static::OPTION_STYLE];
    if ($style) {
      $styled_image_uri = $style->buildUri($file->getFileUri());
      $formatter_build['#image_style'] = $style->id();
      if ($options[static::OPTION_DIMENSION] && file_exists($styled_image_uri)) {
        [$image_width, $image_height] = getimagesize($styled_image_uri);
        $formatter_build['#height'] = $image_height;
        $formatter_build['#width'] = $image_width;
      }
    }
  }

  /**
   * Initialize the image class.
   *
   * @param array $formatter_build
   *   The formatter build array.
   * @param array $options
   *   The options.
   */
  protected function initImagesClasses(array &$formatter_build, array $options) {
    if (!empty($options[static::OPTION_CLASS])) {
      $class = is_array($options[static::OPTION_CLASS]) ? $options[static::OPTION_CLASS] : explode(
        ' ',
        $options[static::OPTION_CLASS]
      );

      if (array_key_exists('#attributes', $formatter_build) && $formatter_build['#attributes']['class']) {
        $formatter_build['#attributes']['class'] = array_merge($formatter_build['#attributes']['class'], $class);
      }
      else {
        $formatter_build['#attributes']['class'] = $class;
      }
    }
  }

  /**
   * Add base data to the image formatter.
   *
   * @param array $formatter_build
   *   The image formatter build array.
   */
  protected function initImageTag(array &$formatter_build) {
    $formatter_build['#item_attributes']['class'] = array_key_exists(
      '#attributes',
      $formatter_build
    ) ? $formatter_build['#attributes']['class'] : [];
    $formatter_build['#item_attributes']['alt'] = $formatter_build['#alt'];
  }

  /**
   * Init image attributes.
   *
   * @param array $formatter_build
   *   The image formatter build array.
   * @param array $options
   *   The options.
   */
  protected function initAttributes(array &$formatter_build, array $options) {
    $formatter_build['#item_attributes'] = $options[static::OPTION_ATTRIBUTES] + $formatter_build['#item_attributes'];
  }

  /**
   * Initialize image meta data.
   *
   * @param array $formatter_build
   *   The image formatter build array.
   * @param \Drupal\media\MediaInterface $media
   *   The media.
   * @param array $options
   *   The options.
   *
   * @throws \Drupal\Core\TypedData\Exception\MissingDataException
   */
  protected function initImageMeta(array &$formatter_build, MediaInterface $media, array $options) {
    $image_data = $media->get($options[static::OPTION_FIELD_NAME])
      ->first()
      ->toArray();
    $formatter_build['#alt'] = empty($image_data['alt']) ? $media->label() : $image_data['alt'];

    // Title.
    if (isset($options[static::OPTION_TITLE]) ||
      isset($image_data['title']) && !empty($image_data['title'])) {
      $formatter_build['#item_attributes']['title'] = $options[static::OPTION_TITLE] ?? $image_data['title'];
    }
  }

  /**
   * Initialise src set attributes for images.
   *
   * @param array $formatter_build
   *   The image formatter build array.
   * @param \Drupal\media\Entity\Media $media
   *   The media.
   * @param array $options
   *   The options.
   */
  protected function initSrcSet(array &$formatter_build, Media $media, array $options) {
    $srcset = $this->getSrcSetFromSourceSettings($media, $options, $options);
    if (!empty($srcset)) {
      array_walk($srcset, function (&$value, $key) {
        $value .= ' ' . $key;
      });
      $formatter_build['#item_attributes']['srcset'] = implode(', ', $srcset);
    }
  }

  /**
   * Init rules for formater.
   *
   * @param array $formatter_build
   *   THe build array.
   * @param \Drupal\media\MediaInterface $media
   *   THe media.
   * @param array $options
   *   The options.
   */
  protected function initRules(array &$formatter_build, MediaInterface $media, array $options) {
    if (isset($options[static::OPTION_RULES])) {
      $settings = [
        static::OPTION_STYLE => $options[static::OPTION_STYLE]->id(),
        static::OPTION_RULES => $options[static::OPTION_RULES],
      ];
      // Get the image file id.
      $file_id = $settings['fileId'] ?? $this->getFileIdFromMedia($media, $options);
      $url = $this->imageBuilder->getImageUrlFromSettings($file_id, $settings, [], $options);
      $formatter_build['#item_attributes'][static::OPTION_OVERRIDE_SRC] = $url;
    }
  }

  /**
   * Return the media url.
   *
   * @param mixed $media_params
   *   THe media.
   * @param array $options
   *   The otpions.
   *
   * @return array
   *   The media Urls.
   */
  public function getMediaUrl($media_params, array $options = []): array {
    // Définitions des options par défaut.
    $options = $this->getDefaultOptions($options);

    // Pour chaque média, on retourne le build array.
    $urls = [];
    foreach ($this->getMediaEntityFromTwigParam($media_params) as $media) {
      $urls[$media->id()] = $this->getStyledUrlFromFileId(
        $this->getFileIdFromMedia($media, $options),
        $options[static::OPTION_STYLE],
        $options[static::OPTION_RELATIVE]
      );
    }

    return $urls;
  }

  /**
   * Return the styled url from file Id.
   *
   * @param string|int $file_id
   *   THe file id.
   * @param mixed $style
   *   The style.
   * @param bool $relative
   *   THe url must be relative or not.
   *
   * @return string|null
   *   The file url.
   */
  public function getStyledUrlFromFileId($file_id, $style = NULL, $relative = TRUE): ?string {
    if ($style) {
      $style = $style instanceof ImageStyleInterface ? $style : $this->conf->getImageStyleStorage()
        ->load($style);
    }
    $url = NULL;
    /** @var \Drupal\file\Entity\File $file */
    $file = $this->conf->getFileStorage()->load($file_id);
    if ($file) {
      if ($style instanceof ImageStyleInterface) {
        $url = $this->fileUrlGenerator->generateAbsoluteString($style->buildUri($file->getFileUri()));
      }
      else {
        $url = $this->fileUrlGenerator->generateAbsoluteString($file->getFileUri());
      }
    }

    if ($relative) {
      $url = $this->fileUrlGenerator->transformRelative($url);
    }

    return $url;
  }

  /**
   * Return the file id from media.
   *
   * @param \Drupal\media\MediaInterface $media
   *   The media.
   * @param array $options
   *   The options.
   *
   * @return string
   *   The file id.
   */
  public function getFileIdFromMedia(MediaInterface $media, array $options): string {
    return $media->get($options[static::OPTION_FIELD_NAME])->target_id;
  }

  /**
   * Build picture tag from media.
   *
   * @param mixed $media_params
   *   The media params from twig.
   * @param array $options
   *   The options.
   *
   * @return array
   *   The build array.
   */
  public function pictureFromMedia($media_params, array $options): array {
    $options = $this->getDefaultOptions($options);
    $picture = [];
    foreach ($this->getMediaEntityFromTwigParam($media_params) as $media) {
      $picture[] = $this->getPictureData($media, $options);
    }

    return $picture;
  }

  /**
   * Return picture data.
   *
   * @param \Drupal\media\MediaInterface $media
   *   The media.
   * @param array $options
   *   The options.
   *
   * @return array
   *   The build array.
   */
  protected function getPictureData(MediaInterface $media, array $options): array {
    // Init the build array.
    $build = [
      '#theme' => 'picture',
    ];

    // Init main attributes.
    $this->initImagesClasses($build, $options);

    // Init main image tag.
    $build['img'] = $this->imgFromMedia($media, $options);

    // Init sources.
    $build['sources'] = [];
    foreach ($options[static::OPTION_SETTINGS] as $mediaQuery => $settings) {
      $build['sources'][] = $this->getSourceFromSettings($media, $mediaQuery, $settings, $options);
    }

    return $build;
  }

  /**
   * Return the source from settings.
   *
   * @param \Drupal\media\MediaInterface $media
   *   The media.
   * @param string $media_query
   *   The media query for the source.
   * @param array $settings
   *   The sources settings.
   * @param array $options
   *   The options.
   *
   * @return array
   *   The source build array.
   */
  protected function getSourceFromSettings(MediaInterface $media, string $media_query, array $settings, array $options): array {
    $src_media = $this->getMediaEntityFromTwigParam($settings['media'] ?? $media);

    return [
      '#theme' => 'source',
      '#settings' => $settings,
      '#options' => $options,
      '#media_query' => $media_query,
      '#src_set' => $this->getSrcSetFromSourceSettings(reset($src_media), $settings, $options),
    ];
  }

  /**
   * Return src set from source settings.
   *
   * @param \Drupal\media\Entity\Media $media
   *   The media.
   * @param array $settings
   *   The src settings.
   * @param array $options
   *   The options.
   *
   * @return array
   *   The list of srcs.
   */
  protected function getSrcSetFromSourceSettings(Media $media, array $settings, array $options): array {
    // Get the image file id.
    $file_id = $settings['fileId'] ?? $this->getFileIdFromMedia($media, $options);
    $default_url = $this->getStyledUrlFromFileId($file_id, $settings[static::OPTION_STYLE] ?? NULL, $options[static::OPTION_RELATIVE]);

    // Build src urls.
    $srcset = [];
    if (isset($settings['srcset'])) {
      $check_oversize_settings = isset($settings['oversize']) ? !$settings['oversize'] : $this->conf->isOversizeCheckEnable();
      if ($check_oversize_settings) {
        $original_file = $this->entityTypeManager->getStorage('file')->load($file_id);
        $settings['srcset'] = $this->oversizeChecker->getOversizeFilteredSettings($original_file, $settings['srcset']);
      }

      foreach ($settings['srcset'] as $type => $data) {
        // Build srcset.
        if ($data === 'default') {
          $srcset[$type] = $default_url;
        }
        elseif ($setUrl = $this->imageBuilder->getImageUrlFromSettings($file_id, $data, $settings, $options)) {
          $srcset[$type] = $setUrl;
        }
      }
    }

    return $srcset;
  }

  /**
   * Return the list of built files from the media.
   *
   * @param \Drupal\media\MediaInterface $media
   *   The media.
   *
   * @return array
   *   The files path.
   */
  public function getBuiltFilesFromMedia(MediaInterface $media): array {
    $file_paths = [];

    // Get file ids.
    foreach ($media->toArray() as $field_name => $value) {
      if (strpos($field_name, 'field_') === 0 && $media->get($field_name) instanceof FileFieldItemList) {
        $file_id = $media->get($field_name)->target_id;
        foreach (['public://', 'private://'] as $dir) {
          try {
            $files = array_filter($this->fileSystem->scanDirectory($dir, '/^' . $file_id . '_/'),
              function ($file) {
                return strpos($file->uri, $this->conf->getDirname()) !== FALSE;
              });
            $file_paths = array_merge($files, $file_paths);
          }
          catch (\Exception $e) {
            // Mute exception...
          }
        }
      }
    }

    return $file_paths;
  }

  /**
   * Clear built files.
   *
   * @param \Drupal\media\MediaInterface $media
   *   Media.
   */
  public function clearBuiltFilesFromMedia(MediaInterface $media): void {
    foreach ($this->getBuiltFilesFromMedia($media) as $file) {
      $this->fileSystem->delete($file->uri);
    }
  }

}
